/*
 * Created on Mar 4, 2004
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */

import java.util.*;

/**
 * @author cslabuser
 *
 * To change the template for this generated type comment go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */

public class CBoard {

	/*
	 * DEFINE VARIABLE
	 */
	public final int NO = 0;
	public final int WHITE_WON = 1;
	public final int BLACK_WON = 2;
	
	/*
	 * DEBUG VARIABLE
	 */
	boolean DEBUG	= false;
		
	// Black is the Color that starting at the top
	/* Whose turn is it? */
	private boolean BlacksMove;
	private boolean	GameEnded = false;
	
	// Boouuuuhhhh this variable is declared public ;]
	public int minmax = 0;
	public String HowDidIGetHere;
	
	/* What is on the square? */
	private int Empty;
	private int Black;
	private int Piece;
	
	/* exclusively for pretty fast bit shifting */
	int bitmask;
	
	/* Access our bit shifting ;o) */
	private void setEmpty(int Square, boolean Value) { if (isEmpty(Square) != Value) { bitmask = 1; bitmask <<= (Square-1); Empty ^= bitmask; } }
	private void setBlack(int Square, boolean Value) { if (isBlack(Square) != Value) { bitmask = 1; bitmask <<= (Square-1); Black ^= bitmask; } }
	private void setPiece(int Square, boolean Value) { if (isPiece(Square) != Value) { bitmask = 1; bitmask <<= (Square-1); Piece ^= bitmask; } }

	/* Access square attributes */
	public boolean isBlacksMove () {return BlacksMove; }
	public boolean isEmpty (int square) { bitmask = Empty; bitmask >>= (square-1); if ((bitmask & 1) == 1) return true; return false; }
	public boolean isBlack (int square) { bitmask = Black; bitmask >>= (square-1); if ((bitmask & 1) == 1) return true; return false; }
	public boolean isPiece (int square) { bitmask = Piece; bitmask >>= (square-1); if ((bitmask & 1) == 1) return true; return false; }
	public boolean hasGameEnded(){return GameEnded;	}
	
	/*
	 * printWhoseTurnItIs : Print who is going to play next: black or white
	 */
	public void printWhoseTurnItIs()
	{
		if(!DEBUG) return; 
		if(BlacksMove)
			System.out.println("It is Black's turn");
		else
			System.out.println("It is White's turn");
	}
	
	/* Constructor, sets the player to begin */
	public CBoard (boolean NewGame, boolean BlackStarts) {

		if (NewGame) {
				
			BlacksMove = BlackStarts;
			
			/* Middle part fields (13-20) are empty */
			for (int i = 13; i < 21; i++) {
				setEmpty(i, true);
			}

			/* Black player occupies the top part (1-12) */	
			for (int i = 1; i < 13; i++) {
				setBlack(i, true);
				setPiece(i, true);
			}

			/* White player occupies the bottom part (21-32) */
			for (int i = 21; i < 33; i++) {
				setBlack(i, false);
				setPiece(i, true);
			}
			
		}
		
	}
	
	/*
	 * CBoard : Constructor. Builds a new board from the given representation (b, B, w, W)
	 */
	public CBoard (char[] Board, boolean BlacksStarts) {
		BlacksMove = BlacksStarts;

		for(int a=1;a<33;a++)
		{
			switch(Board[a-1])
			{
				case 'b':
					setEmpty(a, false);
					setPiece(a, true);
					setBlack(a, true);			
					break;
				case 'B':
					setEmpty(a, false);
					setPiece(a, false);
					setBlack(a, true);
					break;
				case 'w':
					setEmpty(a, false);
					setPiece(a, true);
					setBlack(a, false);				
					break;
				case 'W':
					setEmpty(a, false);
					setPiece(a, false);
					setBlack(a, false);				
					break;
				case ' ':
					setEmpty(a, true);
					// Arbitrary fills the arrays since it's empty
					setPiece(a, false);
					setBlack(a, false);
					break;
			}
		}
	}
	
	public void SwitchTurn() {
		BlacksMove = !BlacksMove;
	}
	
	/* Come on and do some genetics ;o) */
	public CBoard Clone() {

		CBoard Cloned = new CBoard(false, true);
		
		Cloned.Empty = Empty;
		Cloned.Black = Black;
		Cloned.Piece = Piece;
		
		Cloned.BlacksMove = BlacksMove;
		Cloned.GameEnded = GameEnded;
		
		return Cloned;
	}
	
	
	/*
	 * NoPlayerStonLeft : If one player has no more stones, returns true
	 */
	private boolean BlackHasStones(){
		/* This works in such a way, that we go through the array just once,
		 * and when we find both stones (99% of cases), exit the loop */

		for(int a = 1; a < 33; a++) {
			if (!isEmpty(a) && isBlack(a)) return true;
		}
		
		return false;
	}
	private boolean WhiteHasStones(){
		/* This works in such a way, that we go through the array just once,
		 * and when we find both stones (99% of cases), exit the loop */

		for(int a = 32; a > 0; a--) {
			if (!isEmpty(a) && !isBlack(a)) return true;
		}
	
		return false;
	}
	
	/*
	 * MakeMove : Make new move based on input parameter
	 */
	public int MakeMove (String NewMove, boolean CheckValidity) {
		
		int TempSquare1 = 0;
		int TempSquare2 = 0;
		
		/* Firstly if required, check if the move is valid */
		if (CheckValidity) if (!IsMoveLegal(NewMove)) return -1;
		
		/* Move is valid, let's go and update the board */
		StringTokenizer SplittedMoves = new StringTokenizer(NewMove, "-");
		
		/* Get initial position */
		TempSquare1 = Integer.parseInt(SplittedMoves.nextToken());
		
		/* Is it piece or king? */
		boolean FigureType = isPiece(TempSquare1);
				
		/* Remove piece or king from the initial square */
		setEmpty(TempSquare1, true); 
		
		/* Go through all moves and/or jumps and update all squares */
		while (SplittedMoves.hasMoreTokens()) {
			
			TempSquare2 = Integer.parseInt(SplittedMoves.nextToken());
			
			/* If there was a jumped square remove that piece or king as well */
			int JumpedSquare;
			if ((JumpedSquare = GetJumpedSquare(TempSquare1, TempSquare2))>=0) {
				setEmpty(JumpedSquare, true);
			}
			
			TempSquare1 = TempSquare2;
						
		}
		
		/* Place the piece or king to the final square, and if it's a King's row, transform the piece */ 
		setEmpty(TempSquare2, false);
		setBlack(TempSquare2, BlacksMove);
		setPiece(TempSquare2, (isKingsRow(TempSquare2) ? false : FigureType));
		
		if(TheGameIsOver() != NO)
		{
			//GameEnded = true;
			return 2;	// Move correct, but no more pieces, game is over
		}
		
		SwitchTurn();
						
		return 1;
				
	}
	
	/*
	 * GetJumpedSquare : If any square has been jumped, return true and its number
	 */
	private int GetJumpedSquare (int FirstSquare, int SecondSquare) {
		
		/* If any field was jumped, new position must be two squares away in any direction */
		
		if (getUpperLeft(getUpperLeft(FirstSquare)) == SecondSquare) {
			return getUpperLeft(FirstSquare);
		}

		if (getUpperRight(getUpperRight(FirstSquare)) == SecondSquare) {
			return getUpperRight(FirstSquare);
		}

		if (getLowerLeft(getLowerLeft(FirstSquare)) == SecondSquare) {
			return getLowerLeft(FirstSquare);
		}

		if (getLowerRight(getLowerRight(FirstSquare)) == SecondSquare) {
			return getLowerRight(FirstSquare);
		}

		return -1;
		
	}
	
	/*
	 * TheGameIsOver : Check if the game is over
	 */
	public int TheGameIsOver()
	{
		// First check if no stones on the board
		if(BlackHasStones() && !WhiteHasStones())
			return WHITE_WON;
		else if(!BlackHasStones() && WhiteHasStones())
			return BLACK_WON;
			

		// Else check for possible moves, if none, game is over			
		ArrayList LegalMoves = GetLegalMoves();
		if(LegalMoves.isEmpty())
		{
			if(BlacksMove)
				return BLACK_WON;
			else
				return WHITE_WON;
				
			//System.out.println("The Game is over, Legalmoves empty");
			
		}
		else
		{
			//System.out.println("The Game is over, Legalmoves NOT empty");
			return NO;
		}
	}
	
	/*
	 * IsMoveLegal : Check if the move on input is valid
	 */
	private boolean IsMoveLegal (String Move) {
	
		/* Get all legal moves for this turn */
		ArrayList LegalMoves = GetLegalMoves();

		/* If it's valid, we should find it in the list */
		for (int i = 0; i < LegalMoves.size(); i++) {
			if (Move.equals(LegalMoves.get(i)))	return true;
			
		}
		
		return false;
	}
	
	/*
	 * Returns the number of whites stones (pieces or kings)
	 */
	public int NumOfWhiteStones(){
		int a;
		int whites = 0;
		for(a = 0; a<32;a++)
		{
			if(!isEmpty(a) && !isBlack(a))
				whites++;
		}
		return whites;
	}
	
	/*
	 * Returns the number of black stones (pieces or kings)
	 */
	public int NumOfBlackStones(){
		int a;
		int whites = 0;
		for(a = 0; a<32;a++)
		{
			if(!isEmpty(a) && !isBlack(a))
				whites++;
		}
		return whites;
	}
	
	/*
	 * FindJumpList : Works on the MovesList array given as a parameter (9-14-... format)
	 * returns true if the function found something
	 * 			false otherwise
	 */
	private boolean FindJumpList(ArrayList MovesList, String sPos)
	{
		boolean	CurrentPiece = true;
		int 	PreviousSquare = 0, CurrentSquare = 0;
		int 	TempSquare = 0;
		boolean FoundJumps = false;
		
		String 	MovesSplitted = new String(sPos);
		
		// Tokenize the string to get the moves done so far
		StringTokenizer MovesTokenizer = new StringTokenizer(MovesSplitted, "-");
		
		// Loop through all the moves to get the first pos and the last one (which we want to work on)
		while(MovesTokenizer.hasMoreTokens())
		{
			PreviousSquare = CurrentSquare;
			CurrentSquare = Integer.parseInt(MovesTokenizer.nextToken());
			if(PreviousSquare == 0)
			{
				CurrentPiece = isPiece(CurrentSquare);
			}
		}
		
		// We have to check for the jumps in the 4 directions
		if(BlacksMove) // If we have either a BlackMove or a King
		{
			int iLimit = 2;
			if(!CurrentPiece)
				iLimit = 4;

			// We check for the 4 directions if we have a king, only 2 otherwise				
			for(int p=0;p<iLimit;p++)
			{
				switch(p)
				{
					case 0:
						TempSquare = getLowerLeft(CurrentSquare);
						break;
					case 1:
						TempSquare = getLowerRight(CurrentSquare);
						break;
					case 2:
						TempSquare = getUpperLeft(CurrentSquare);
						break;
					case 3:
						TempSquare = getUpperRight(CurrentSquare);
						break;
				}
				
				// Black's Move, look downwards
				if(TempSquare != 0 && !isBlack(TempSquare) && !isEmpty(TempSquare))		// There's a White in the lowerleft
				{
					// Can we jump??
					switch(p)
					{
						case 0:
							TempSquare = getLowerLeft(TempSquare);
							break;
						case 1:
							TempSquare = getLowerRight(TempSquare);
							break;
						case 2:
							TempSquare = getUpperLeft(TempSquare);
							break;
						case 3:
							TempSquare = getUpperRight(TempSquare);
							break;
					}
					
					// We check that we have a valid state 
					if(TempSquare != 0 && TempSquare != PreviousSquare && isEmpty(TempSquare))		// Free pos, so we can jump
					{
						// Check if we are not looping
						String  S = new String(sPos);
						boolean looping = false;				
						// Tokenize the string to get the moves done so far
						StringTokenizer MT = new StringTokenizer(S, "-");
		
						// Loop through all the moves to get the first pos and the last one (which we want to work on)
						while(MT.hasMoreTokens())
						{
							if(TempSquare == Integer.parseInt(MT.nextToken()))
							{
								looping = true;
							}
						}

						if(!looping)
						{
							// Add this move to the list
							String NewList = sPos + "-" + TempSquare;
						
							FoundJumps = true;
						
							// Call the recursion to see forward...
							if(!FindJumpList(MovesList, NewList))
								MovesList.add(NewList);
						}  
					}
				}	
			}
		
		}
		else		// White's turn, or king (so can move in 4 directions)
		{
			int iLimit = 2;
			if(!CurrentPiece)
				iLimit = 4;

			// We check for the 4 directions if we have a king, only 2 otherwise				
			for(int p=0;p<iLimit;p++)
			{
				switch(p)
				{
					case 0:
						TempSquare = getUpperLeft(CurrentSquare);
						break;
					case 1:
						TempSquare = getUpperRight(CurrentSquare);
						break;
					case 2:
						TempSquare = getLowerLeft(CurrentSquare);
						break;
					case 3:
						TempSquare = getLowerRight(CurrentSquare);
						break;
				}
				
				// Black's Move, look downwards
				if(TempSquare != 0 && isBlack(TempSquare) && !isEmpty(TempSquare))		// There's a White in the lowerleft
				{
					// Can we jump??
					switch(p)
					{
						case 0:
							TempSquare = getUpperLeft(TempSquare);
							break;
						case 1:
							TempSquare = getUpperRight(TempSquare);
							break;
						case 2:
							TempSquare = getLowerLeft(TempSquare);
							break;
						case 3:
							TempSquare = getLowerRight(TempSquare);
							break;
					}

					// We check that we are not going back to a previous position
					if(TempSquare != 0 && TempSquare != PreviousSquare && isEmpty(TempSquare))		// Free pos, so we can jump
					{
						// Check if we are not looping
						String  S = new String(sPos);
						boolean looping = false;				
						// Tokenize the string to get the moves done so far
						StringTokenizer MT = new StringTokenizer(S, "-");
		
						// Loop through all the moves to get the first pos and the last one (which we want to work on)
						while(MT.hasMoreTokens())
						{
							if(TempSquare == Integer.parseInt(MT.nextToken()))
							{
								looping = true;
							}
						}

						if(!looping)
						{
							// Add this move to the list
							String NewList = sPos + "-" + TempSquare;
							
							FoundJumps = true;
						
							// Call the recursion to see forward...
							if(!FindJumpList(MovesList, NewList))
								MovesList.add(NewList);
						} 					
					}
				}	
			}
		}
		
		// Returns true if some jumps were found
		return FoundJumps;
	}
	
	/* 
	 * GetLegalMoves : This function gets the legal moves from the given position (moves + jumps)
	 */
	public ArrayList GetLegalMoves()
	{
		ArrayList 	MovesList = new ArrayList();
		int 		iPos = 0;
		boolean		FoundJumps = false;
		boolean 	CanMove = false;
		
		// First we have to check if we have a jump pending
		for(int From=1;From<33;From++)
		{
			// If nothing can move, don't consider this position
			if(isEmpty(From))
				continue;
			else
			{
				if(BlacksMove && !isBlack(From) )			// If it's black's turn and we don't have a black piece, continue
					continue;
				if(!BlacksMove && isBlack(From) )			// If it's white's turn and we don't have a white piece, continue
					continue;			
			}

			if(FindJumpList(MovesList, String.valueOf(From)) == true)
				FoundJumps = true;
		}
		
		// If we found some jumps to do, do them
		if(FoundJumps)
			return MovesList;
		
		// No jumps to do, proceed to next step....
		// Then we check whose turn it is, see if we can move
		if(BlacksMove)	// We go from the top to the bottom
		{
			// Check for all the pieces if they can move down
			for(int a=1;a<33;a++)
			{
				// Non empty square with a black piece				
				if(!isEmpty(a) && isBlack(a))
				{
					int iLimit = 2;
				
					if(!isPiece(a))	// We have a king
						iLimit = 4;	// We check the fourth directions
						
					for(int p=0;p<iLimit;p++)
					{
						switch(p)
						{
							case 0:
								iPos = getLowerLeft(a);
								break;							
							case 1:
								iPos = getLowerRight(a);
								break;							
							case 2:
								iPos = getUpperLeft(a);
								break;							
							case 3:
								iPos = getUpperRight(a);
								break;
						}
						
						// Check for the current square
						if(iPos != 0 && isEmpty(iPos))		// The position is valid (<> 0)
						{
							MovesList.add(new String((a) + "-" + iPos));
							CanMove = true;
						}
							
					}
				}
			}
		}
		else		// We put the same stuff in here, except that we will look for move upwards
		{
			// Check for all the pieces if they can move down
			for(int a=1;a<33;a++)
			{
				// Non empty square with a black piece				
				if(!isEmpty(a) && !isBlack(a))
				{
					int iLimit = 2;
					if(!isPiece(a))	// We have a king
						iLimit = 4;	// We check the fourth directions
						
					for(int p=0;p<iLimit;p++)
					{
						switch(p)
						{
							case 0:
								iPos = getUpperLeft(a);
								break;							
							case 1:
								iPos = getUpperRight(a);
								break;
							case 2:
								iPos = getLowerLeft(a);
								break;							
							case 3:
								iPos = getLowerRight(a);
								break;							
						}
						
						// Check for the current square
						if(iPos != 0 && isEmpty(iPos))		// The position is valid (<> 0)
						{
							MovesList.add(new String((a) + "-" + iPos));
							CanMove = true;
						}
							
					}
				}
			}
			
		}
		
		if(!CanMove) {
			/*
			System.out.println(Display());
			if (MovesList.size()==0) System.out.println("No things in MoveList");
			if (NoPlayersStonesLeft()) System.out.println("NoPlayersStonesLeft"); else System.out.println("PlayersStonesLeft!!!");
			if (BlacksMove) System.out.println("BlacksMove"); else System.out.println("WhitesMove");
			System.out.println("No move possible. Game is over.");
			*/
		}
		
		// Return the list of possible moves
		return MovesList;
	}

	/*
	 * isKingsRow : Is this the King's Row??
	 */
	public boolean isKingsRow(int square)
	{
		return ((isBlacksMove() && (square >= 29)) || !isBlacksMove() && (square <= 4));
	}

	/*
	Board:
	*    1   2   3   4
	*  5   6   7   8
	*    9  10  11  12
	* 13  14  15  16
	*   17  18  19  20
	* 21  22  23  24
	*   25  26  27  28
	* 29  30  31  32
	*/
	//	Which squares are around this one ?
	// Returns 0 if no error
	private int getUpperLeft(int Pos)
	{
		switch(Pos)
		{
			// We have 3 different cases:
			// The even rows, the odd rows, and the on-the-boarder squares 
			
			// ==> even rows
			case 6: 	case 7: 	case 8:	
			case 14: 	case 15: 	case 16:
			case 22: 	case 23: 	case 24:
			case 30: 	case 31: 	case 32:
				return Pos -  5;
			// ==> odd rows
			case 9: 	case 10: 	case 11:	case 12:	
			case 17: 	case 18: 	case 19:	case 20:
			case 25: 	case 26: 	case 27:	case 28:			
				return Pos -  4;
			default:	return 0;	

		}
	}

	/*
	Board:
	*    1   2   3   4
	*  5   6   7   8
	*    9  10  11  12
	* 13  14  15  16
	*   17  18  19  20
	* 21  22  23  24
	*   25  26  27  28
	* 29  30  31  32
	*/
	private int getUpperRight(int Pos)
	{
		switch(Pos)
		{
			// We have 3 different cases:
			// The even rows, the odd rows, and the on-the-boarder squares 
			
			// ==> even rows
			case 5: 	case 6: 	case 7:		case 8:	
			case 13: 	case 14: 	case 15:	case 16:
			case 21: 	case 22: 	case 23:	case 24:
			case 29: 	case 30: 	case 31:	case 32:						
				return Pos -  4;
			// ==> odd rows
			case 9: 	case 10: 	case 11:	
			case 17: 	case 18: 	case 19:
			case 25: 	case 26: 	case 27:			
				return Pos -  3;
			default:	return 0;	
		}
	}

	/*
	Board:
	*    1   2   3   4
	*  5   6   7   8
	*    9  10  11  12
	* 13  14  15  16
	*   17  18  19  20
	* 21  22  23  24
	*   25  26  27  28
	* 29  30  31  32
	*/
	private int getLowerLeft(int Pos)
	{
		switch(Pos)
		{
			// We have 3 different cases:
			// The even rows, the odd rows, and the on-the-boarder squares 
			
			// ==> odd rows
			case 1: 	case 2: 	case 3:		case 4:	
			case 9: 	case 10: 	case 11:	case 12:
			case 17: 	case 18: 	case 19:	case 20:
			case 25: 	case 26: 	case 27:	case 28:						
				return Pos + 4;
			// ==> even rows
			case 6: 	case 7: 	case 8:	
			case 14: 	case 15: 	case 16:
			case 22: 	case 23: 	case 24:			
				return Pos + 3;
			default:	return 0;	
		}
	}

	/*
	Board:
	*    1   2   3   4
	*  5   6   7   8
	*    9  10  11  12
	* 13  14  15  16
	*   17  18  19  20
	* 21  22  23  24
	*   25  26  27  28
	* 29  30  31  32
	*/
	//	Which squares are around this one ?
	// Returns 0 if no error
	private int getLowerRight(int Pos)
	{
		switch(Pos)
		{
			// We have 3 different cases:
			// The even rows, the odd rows, and the on-the-boarder squares 
			
			// ==> even rows
			case 5:		case 6: 	case 7: 	case 8:	
			case 13:	case 14: 	case 15: 	case 16:
			case 21:	case 22: 	case 23: 	case 24:
				return Pos + 4;
			// ==> odd rows
			case 1: 	case 2: 	case 3:
			case 9: 	case 10: 	case 11:	
			case 17: 	case 18: 	case 19:
			case 25: 	case 26: 	case 27:			
				return Pos + 5;
			default:	return 0;	

		}
	}
	
	/*
	 * Display : Displays the board in the console
	 */
	public String Display() {
		
		String CurrentBoard = new String();
	
		for(int a = 0; a < 32; a++)
		{
			if(a%4 == 0 && a!=0 && a!=32)
				CurrentBoard += "\n";
			if(a%8 == 0)
				CurrentBoard += "   ";
			if(a%4 != 0)
				CurrentBoard += "     ";
				
			// Is there something to display?
			if(!isEmpty(a+1))
			{
				if(isPiece(a+1) && isBlack(a+1))		// We have a black piece
					CurrentBoard += "b";
				else if (!isPiece(a+1) && isBlack(a+1))		// We have a black King
					CurrentBoard += "B";
				else if(isPiece(a+1) && !isBlack(a+1))		// We have a white piece
					CurrentBoard += "w";
				else if(!isPiece(a+1) && !isBlack(a+1))		// We have a white King
					CurrentBoard += "W";
				
			}
			else 	// Nothing to display
				CurrentBoard += "";
		}
		CurrentBoard += "\n";
		
		return CurrentBoard;
	}

}


